/*
 * packet-sstp.h
 * Routines for SSTP dissection
 * by Christophe Alladoum <christophe.alladoum@hsc.fr>
 * under GPL v2
 * refs :
 * http://www.wireshark.org/docs/wsdg_html_chunked/
 * http://anonsvn.wireshark.org/wireshark/trunk/doc/README.developer 
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <glib.h>
#include <epan/packet.h>
#include <epan/dissectors/packet-ssl.h>

/* #define MIN(a,b) ((a<b) ? a : b) */

/* SSTP Properties */
#define SSTP_MIN_LEN 4 
#define SSL_DEFAULT_RANGE "443"
#define SSTP_MAX_ATTR 256

/* SSTP Packet Type */
#define SSTP_DATA_PACKET 0x00
#define SSTP_CONTROL_PACKET 0x01

/* SSTP Protocol Type */
#define SSTP_ENCAPSULATED_PROTOCOL_PPP 0x0001

/* SSTP Status Message */
#define SSTP_MSG_CALL_CONNECT_REQUEST 0x0001
#define SSTP_MSG_CALL_CONNECT_ACK 0x0002
#define SSTP_MSG_CALL_CONNECT_NAK 0x0003
#define SSTP_MSG_CALL_CONNECTED 0x0004
#define SSTP_MSG_CALL_ABORT 0x0005
#define SSTP_MSG_CALL_DISCONNECT 0x0006
#define SSTP_MSG_CALL_DISCONNECT_ACK 0x0007
#define SSTP_MSG_ECHO_REQUEST 0x0008
#define SSTP_MSG_ECHO_RESPONSE 0x0009

/* SSTP Attribute Message Type */
#define SSTP_ATTRIB_NO_ERROR 0x00
#define SSTP_ATTRIB_ENCAPSULATED_PROTOCOL_ID 0x01
#define SSTP_ATTRIB_STATUS_INFO 0x02
#define SSTP_ATTRIB_CRYPTO_BINDING 0x03
#define SSTP_ATTRIB_CRYPTO_BINDING_REQ 0x04

/* Crypto Binding Request Attribute */
#define CERT_HASH_PROTOCOL_SHA1 0x01
#define CERT_HASH_PROTOCOL_SHA256 0x02

/* Status Info Attribute */
#define ATTRIB_STATUS_NO_ERROR 0x00000000
#define ATTRIB_STATUS_DUPLICATE_ATTRIBUTE 0x00000001
#define ATTRIB_STATUS_UNRECOGNIZED_ATTRIBUTE 0x00000002
#define ATTRIB_STATUS_INVALID_ATTRIB_VALUE_LENGTH 0x00000003
#define ATTRIB_STATUS_VALUE_NOT_SUPPORTED 0x00000004
#define ATTRIB_STATUS_UNACCEPTED_FRAME_RECEIVED 0x00000005
#define ATTRIB_STATUS_RETRY_COUNT_EXCEEDED 0x00000006
#define ATTRIB_STATUS_INVALID_FRAME_RECEIVED 0x00000007
#define ATTRIB_STATUS_NEGOTIATION_TIMEOUT 0x00000008
#define ATTRIB_STATUS_ATTRIB_NOT_SUPPORTED_IN_MSG 0x00000009
#define ATTRIB_STATUS_REQUIRED_ATTRIBUTE_MISSING 0x0000000a
#define ATTRIB_STATUS_STATUS_INFO_NOT_SUPPORTED_IN_MSG 0x0000000b


/* VALS() structure definitions */
static const value_string packet_types[] =
  {
    {SSTP_DATA_PACKET, "SSTP Data Packet"},
    {SSTP_CONTROL_PACKET, "SSTP Control Packet"},
    {0, NULL }
  };

static const value_string message_types[] =
  {
    {SSTP_MSG_CALL_CONNECT_REQUEST, "SSTP_MSG_CALL_CONNECT_REQUEST"},
    {SSTP_MSG_CALL_CONNECT_ACK, "SSTP_MSG_CALL_CONNECT_ACK"},
    {SSTP_MSG_CALL_CONNECT_NAK, "SSTP_MSG_CALL_CONNECT_NAK"},
    {SSTP_MSG_CALL_CONNECTED, "SSTP_MSG_CALL_CONNECTED"},
    {SSTP_MSG_CALL_ABORT, "SSTP_MSG_CALL_ABORT"},
    {SSTP_MSG_CALL_DISCONNECT, "SSTP_MSG_CALL_DISCONNECT"},
    {SSTP_MSG_CALL_DISCONNECT_ACK, "SSTP_MSG_CALL_DISCONNECT_ACK"},
    {SSTP_MSG_ECHO_REQUEST, "SSTP_MSG_ECHO_REQUEST" },
    {SSTP_MSG_ECHO_RESPONSE, "SSTP_MSG_ECHO_RESPONSE" },
    {0, NULL }
  };

static const value_string attributes_ids[] =
  {
    {SSTP_ATTRIB_NO_ERROR, "SSTP_ATTRIB_NO_ERROR"},
    {SSTP_ATTRIB_ENCAPSULATED_PROTOCOL_ID, "SSTP_ATTRIB_ENCAPSULATED_PROTOCOL_ID"},
    {SSTP_ATTRIB_STATUS_INFO, "SSTP_ATTRIB_STATUS_INFO"},
    {SSTP_ATTRIB_CRYPTO_BINDING, "SSTP_ATTRIB_CRYPTO_BINDING"},
    {SSTP_ATTRIB_CRYPTO_BINDING_REQ, "SSTP_ATTRIB_CRYPTO_BINDING_REQ"},
    {0, NULL }
  };

static const value_string encapsulated_protocol_id_ids[] =
  {
    { SSTP_ENCAPSULATED_PROTOCOL_PPP, "SSTP_ENCAPSULATED_PROTOCOL_PPP" },
    {0, NULL }
  };

static const value_string hash_protocol_bitmasks[] = 
  {
    {CERT_HASH_PROTOCOL_SHA1, "CERT_HASH_PROTOCOL_SHA1"}, 
    {CERT_HASH_PROTOCOL_SHA256, "CERT_HASH_PROTOCOL_SHA256"},
    {0, NULL }    
  };
      
static const value_string status_info[] =
  {
    {ATTRIB_STATUS_NO_ERROR, "ATTRIB_STATUS_NO_ERROR"},
    {ATTRIB_STATUS_DUPLICATE_ATTRIBUTE, "ATTRIB_STATUS_DUPLICATE_ATTRIBUTE"},
    {ATTRIB_STATUS_UNRECOGNIZED_ATTRIBUTE, "ATTRIB_STATUS_UNRECOGNIZED_ATTRIBUTE"},
    {ATTRIB_STATUS_INVALID_ATTRIB_VALUE_LENGTH, "ATTRIB_STATUS_INVALID_ATTRIB_VALUE_LENGTH"},
    {ATTRIB_STATUS_VALUE_NOT_SUPPORTED, "ATTRIB_STATUS_VALUE_NOT_SUPPORTED"},
    {ATTRIB_STATUS_UNACCEPTED_FRAME_RECEIVED, "ATTRIB_STATUS_UNACCEPTED_FRAME_RECEIVED"}, 
    {ATTRIB_STATUS_RETRY_COUNT_EXCEEDED, "ATTRIB_STATUS_RETRY_COUNT_EXCEEDED"},
    {ATTRIB_STATUS_INVALID_FRAME_RECEIVED, "ATTRIB_STATUS_INVALID_FRAME_RECEIVED"},
    {ATTRIB_STATUS_NEGOTIATION_TIMEOUT, "ATTRIB_STATUS_NEGOTIATION_TIMEOUT"},
    {ATTRIB_STATUS_ATTRIB_NOT_SUPPORTED_IN_MSG, "ATTRIB_STATUS_ATTRIB_NOT_SUPPORTED_IN_MSG"},
    {ATTRIB_STATUS_REQUIRED_ATTRIBUTE_MISSING, "ATTRIB_STATUS_REQUIRED_ATTRIBUTE_MISSING"}, 
    {ATTRIB_STATUS_STATUS_INFO_NOT_SUPPORTED_IN_MSG, "ATTRIB_STATUS_STATUS_INFO_NOT_SUPPORTED_IN_MSG"},
    {0, NULL }
  };

/* Functions declaration */
void proto_reg_handoff_sstp(void);
void proto_register_sstp(void);
gboolean is_control_message(tvbuff_t *tvb);
